package com.supervision.pdfqaserver.service.impl;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.json.JSONUtil;
import com.supervision.pdfqaserver.cache.PromptCache;
import com.supervision.pdfqaserver.dao.Neo4jRepository;
import com.supervision.pdfqaserver.domain.Intention;
import com.supervision.pdfqaserver.dto.*;
import com.supervision.pdfqaserver.dto.neo4j.NodeDTO;
import com.supervision.pdfqaserver.dto.neo4j.PathDTO;
import com.supervision.pdfqaserver.dto.neo4j.RelationshipValueDTO;
import com.supervision.pdfqaserver.service.AiCallService;
import com.supervision.pdfqaserver.service.DomainMetadataService;
import com.supervision.pdfqaserver.service.IntentionService;
import com.supervision.pdfqaserver.service.TripleToCypherExecutor;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.neo4j.driver.Record;
import org.neo4j.driver.internal.InternalNode;
import org.neo4j.driver.internal.InternalRelationship;
import org.springframework.stereotype.Service;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import static com.supervision.pdfqaserver.cache.PromptCache.*;

@Slf4j
@Service
@RequiredArgsConstructor
public class TripleToCypherExecutorImpl implements TripleToCypherExecutor {

    private final Neo4jRepository neo4jRepository;

    private final IntentionService intentionService;
    private static volatile CypherSchemaDTO cypherSchemaDTO;

    private final AiCallService aiCallService;

    private final DomainMetadataService domainMetadataService;

    @Override
    public String generateInsertCypher(EREDTO eredto) {

        String prompt = PromptCache.promptMap.get(ERE_TO_INSERT_CYPHER);
        return aiCallService.call(prompt);
    }

    @Override
    public String generateQueryCypher(String query,String domainCategoryId) {
        List<Intention> intentions = intentionService.listAllPassed();
        List<Intention> relations = classifyIntents(query, intentions);
        if (CollUtil.isEmpty(relations)) {
            log.info("没有找到匹配的意图，query: {}", query);
            return null;
        }
        List<DomainMetadataDTO> domainMetadataDTOS = domainMetadataService.listByIntentionIds(relations.stream().map(Intention::getId).toList());
        CypherSchemaDTO schemaDTO = this.queryRelationSchema(domainMetadataDTOS);
        if (CollUtil.isEmpty(schemaDTO.getRelations()) && CollUtil.isEmpty(schemaDTO.getNodes())) {
            log.info("没有找到匹配的关系或实体，query: {}", query);
            return null;
        }
        String prompt = promptMap.get(TEXT_TO_CYPHER_2);
        String format = StrUtil.format(prompt, Map.of("question", query, "schema", schemaDTO.format()));
        String call = aiCallService.call(format);
        if (StrUtil.equals(call,"I could not generate a Cypher script; the required information is not part of the Neo4j schema.")){
            log.info("大模型没能生成cypher，query: {}", query);
            return null;
        }
        return call;
    }

    @Override
    public List<Map<String, Object>> executeCypher(String cypher) {
        List<Record> records = neo4jRepository.executeCypherNative(cypher, null);
        return mapRecords(records);
    }


    private List<Map<String, Object>> mapRecords(List<Record> records) {
        List<Map<String, Object>> recordList = new ArrayList<>();
        for (Record record : records) {
            HashMap<String, Object> map = new HashMap<>();
            for (String key : record.keys()) {
                org.neo4j.driver.Value value = record.get(key);
                String typeName = value.type().name();
                if (typeName.equals("NULL")){
                    map.put(key,null);
                }

                if (StrUtil.equalsAny(typeName, "BOOLEAN","STRING", "NUMBER", "INTEGER", "FLOAT")){
                    // MATCH (n)-[r]-() where n.caseId= '1'  RETURN  n.recordId limit 10
                    map.put(key,value.asObject());
                }

                if (typeName.equals("PATH")){
                    // MATCH p=(n)-[*2]-() where n.caseId= '1'  RETURN  p limit 10
                    map.put(key,new PathDTO(value.asPath()));
                }

                if (typeName.equals("RELATIONSHIP")){
                    // MATCH (n)-[r]-() where n.caseId= '1'  RETURN  r limit 10
                    map.put(key,new RelationshipValueDTO((InternalRelationship) value.asRelationship()));
                }
                if (typeName.equals("LIST OF ANY?")){

                    List<RelationshipValueDTO> list = value.asList().stream()
                            .map(i -> new RelationshipValueDTO((InternalRelationship) i)).toList();
                    map.put(key,list);
                }
                if (typeName.equals("NODE")){
                    // MATCH (n)-[r]-() where n.caseId= '1'  RETURN  r limit 10
                    map.put(key,new NodeDTO((InternalNode) value.asNode()));
                }
                recordList.add(map);
            }
        }
        return recordList;
    }

    @Override
    public void saveERE(EREDTO eredto) {

        List<EntityExtractionDTO> entities = eredto.getEntities();
        Map<String, List<Long>> nodeCache = new HashMap<>();
        if (CollUtil.isNotEmpty(entities)){
            // 保存节点
            for (EntityExtractionDTO entity : entities) {
                if (StrUtil.isEmpty(entity.getName())){
                    log.info("实体name属性为空,详情:{}", JSONUtil.toJsonStr(entity));
                    continue;
                }
                Map<String, Object> attributes = entity.getAttributes().stream().collect(Collectors.toMap(
                        TruncationERAttributeDTO::getAttributeEn, TruncationERAttributeDTO::getValue
                ));
                attributes.put("truncationId", entity.getTruncationId());
                attributes.put("name", entity.getName());
                log.info("保存节点{},属性:{}", entity.getEntityEn(),JSONUtil.toJsonStr(entity.getAttributes()));
                List<Long> nodeIds = neo4jRepository.saveOrUpdateEntityNode(entity.getEntityEn(), "name", attributes);
                nodeCache.put(StrUtil.join("_", entity.getEntity(), entity.getName()), nodeIds);
            }
        }
        if (CollUtil.isNotEmpty(eredto.getRelations())){
            // 保存关系
            for (RelationExtractionDTO relation : eredto.getRelations()) {
                String sourceNodeKey = StrUtil.join("_", relation.getSourceType(), relation.getSource());
                List<Long> sourceNodeIds = nodeCache.get(sourceNodeKey);
                if (CollUtil.isEmpty(sourceNodeIds)) {
                    log.info("关系{}没有source节点", sourceNodeKey);
                    continue;
                }
                String targetNodeKey = StrUtil.join("_", relation.getTargetType(), relation.getTarget());
                List<Long> targetNodeIds = nodeCache.get(targetNodeKey);
                if (CollUtil.isEmpty(targetNodeIds)) {
                    log.info("关系{}没有target节点", targetNodeKey);
                    continue;
                }
                Map<String, Object> attributes = relation.getAttributes().stream().collect(Collectors.toMap(
                        TruncationERAttributeDTO::getAttributeEn, TruncationERAttributeDTO::getValue
                ));
                attributes.put("truncationId", relation.getTruncationId());
                for (Long sourceNodeId : sourceNodeIds) {
                    for (Long targetNodeId : targetNodeIds) {
                        if (sourceNodeId.equals(targetNodeId)) {
                            log.info("关系{}的source和target节点相同", sourceNodeKey);
                            continue;
                        }
                        log.info("保存关系{}-{}-{}的属性:{}", relation.getSourceTypeEn(), relation.getRelationEn(),relation.getTargetTypeEn(), attributes);
                        List<Long> longs = neo4jRepository.saveOrUpdateRelation(sourceNodeId, targetNodeId, relation.getRelationEn(), false, false, attributes);
                    }
                }
            }
        }
    }

    @Override
    public CypherSchemaDTO loadGraphSchema() {

        List<RelationExtractionDTO> relationSchema = neo4jRepository.getRelationSchema();
        List<EntityExtractionDTO> entitySchema = neo4jRepository.getNodeSchema();
        return new CypherSchemaDTO(entitySchema, relationSchema);
    }

    @Override
    public CypherSchemaDTO queryRelationSchema(List<DomainMetadataDTO> metadataDTOS) {
        if (CollUtil.isEmpty(metadataDTOS)){
            return null;
        }
        if (cypherSchemaDTO == null) {
            synchronized (TripleToCypherExecutorImpl.class) {
                if (cypherSchemaDTO == null) {
                    cypherSchemaDTO = this.loadGraphSchema();
                }
            }
        }
        List<RelationExtractionDTO> merged = new ArrayList<>();
        for (DomainMetadataDTO metadataDTO : metadataDTOS) {
            String relation = metadataDTO.getRelation();
            String sourceType = metadataDTO.getSourceType();
            String targetType = metadataDTO.getTargetType();
            if (StrUtil.isEmpty(relation) || StrUtil.isEmpty(sourceType) || StrUtil.isEmpty(targetType)){
                log.warn("元数据中关系、源类型或目标类型为空，无法查询关系schema: {}", metadataDTO);
                continue;
            }

            RelationExtractionDTO rel = cypherSchemaDTO.getRelation(sourceType, relation, targetType);
            if (null == rel){
                continue;
            }
            List<RelationExtractionDTO> relSourceType = cypherSchemaDTO.getRelations(sourceType);
            List<RelationExtractionDTO> relTargetType = cypherSchemaDTO.getRelations(targetType);

            relSourceType.add(rel);
            relSourceType.addAll(relTargetType);
            for (RelationExtractionDTO relationExtractionDTO : relSourceType) {
                boolean none = merged.stream().noneMatch(i -> StrUtil.equals(i.getRelation(), relationExtractionDTO.getRelation()) &&
                        StrUtil.equals(i.getSourceType(), relationExtractionDTO.getSourceType()) &&
                        StrUtil.equals(i.getTargetType(), relationExtractionDTO.getTargetType()));
                if (none){
                    merged.add(relationExtractionDTO);
                }
            }
        }
        List<EntityExtractionDTO> entityExtractionDTOS = new ArrayList<>();
        for (RelationExtractionDTO relationExtractionDTO : merged) {
            EntityExtractionDTO node = cypherSchemaDTO.getNode(relationExtractionDTO.getSourceType());
            if (null != node){
                boolean none = entityExtractionDTOS.stream().noneMatch(
                        entity -> StrUtil.equals(entity.getEntity(), node.getEntity())
                );
                if (none) {
                    entityExtractionDTOS.add(node);
                }
            }
        }
        return new CypherSchemaDTO(
                entityExtractionDTOS,
                merged
        );
    }

    private List<Intention> classifyIntents(String query, List<Intention> intentions) {
        if (StrUtil.isEmpty(query) || CollUtil.isEmpty(intentions)) {
            return new ArrayList<>();
        }
        String prompt = promptMap.get(CLASSIFY_QUERY_INTENT);
        List<Intention> result = new ArrayList<>();
        log.info("开始分类意图，query: {}, intentions size: {}", query, intentions.size());
        List<List<Intention>> intentionSplit = CollUtil.split(intentions, 200);
        for (List<Intention> intentionList : intentionSplit) {
            log.info("分类意图，query: {}, intentions size: {}", query, intentionList.size());
            String intents = intentionList.stream().map(i -> "  - " + i.getDigest() + "\n").collect(Collectors.joining());
            Map<String, Object> params = Map.of("query", query,
                    "intents", intents);
            String format = StrUtil.format(prompt, params);
            String call = aiCallService.call(format);

            if (StrUtil.isEmpty(call)) {
                return new ArrayList<>();
            }
            List<String> digests = JSONUtil.parseArray(call).stream().map(Object::toString).toList();
            if (CollUtil.isEmpty(digests)) {
                continue;
            }
            List<Intention> collect = intentionList.stream().filter(i -> digests.contains(i.getDigest())).collect(Collectors.toList());
            if (CollUtil.isNotEmpty(collect)) {
                result.addAll(collect);
            }
        }
        return result;
    }
}
